@@ -1,22 +1,16 @@
 import itertools
-import json
 import logging
-import tempfile
 from concurrent.futures import ThreadPoolExecutor, as_completed
 from functools import partial
 from multiprocessing import cpu_count
-from operator import itemgetter
 from urllib.parse import urlparse
 
 from funcy import cached_property
 
-from dvc.exceptions import (
-    DvcException,
-    DvcIgnoreInCollectedDirError,
-    RemoteCacheRequiredError,
-)
+from dvc.exceptions import DvcException, DvcIgnoreInCollectedDirError
+from dvc.hash_info import HashInfo
 from dvc.ignore import DvcIgnore
-from dvc.path_info import PathInfo, URLInfo
+from dvc.path_info import URLInfo
 from dvc.progress import Tqdm
 from dvc.state import StateNoop
 from dvc.utils import tmp_fname
@@ -242,7 +236,7 @@ class BaseTree:
         )
 
         if not self.exists(path_info):
-            return self.PARAM_CHECKSUM, None
+            return None
 
         # pylint: disable=assignment-from-none
         hash_ = self.state.get(path_info)
@@ -260,28 +254,24 @@ class BaseTree:
             hash_ = None
 
         if hash_:
-            return self.PARAM_CHECKSUM, hash_
+            hash_info = HashInfo(self.PARAM_CHECKSUM, hash_)
+            if hash_info.isdir:
+                hash_info.dir_info = self.cache.get_dir_cache(hash_info)
+            return hash_info
 
         if self.isdir(path_info):
-            typ, hash_ = self.get_dir_hash(path_info, **kwargs)
+            hash_info = self.get_dir_hash(path_info, **kwargs)
         else:
-            typ, hash_ = self.get_file_hash(path_info)
+            hash_info = self.get_file_hash(path_info)
 
-        if hash_ and self.exists(path_info):
-            self.state.save(path_info, hash_)
+        if hash_info and self.exists(path_info):
+            self.state.save(path_info, hash_info.value)
 
-        return typ, hash_
+        return hash_info
 
     def get_file_hash(self, path_info):
         raise NotImplementedError
 
-    def get_dir_hash(self, path_info, **kwargs):
-        if not self.cache:
-            raise RemoteCacheRequiredError(path_info)
-
-        dir_info = self._collect_dir(path_info, **kwargs)
-        return self.save_dir_info(dir_info)
-
     def hash_to_path_info(self, hash_):
         return self.path_info / hash_[0:2] / hash_[2:]
 
@@ -293,10 +283,6 @@ class BaseTree:
 
         return "".join(parts)
 
-    def save_info(self, path_info, **kwargs):
-        typ, hash_ = self.get_hash(path_info, **kwargs)
-        return {typ: hash_}
-
     def _calculate_hashes(self, file_infos):
         file_infos = list(file_infos)
         with Tqdm(
@@ -306,9 +292,7 @@ class BaseTree:
         ) as pbar:
             worker = pbar.wrap_fn(self.get_file_hash)
             with ThreadPoolExecutor(max_workers=self.hash_jobs) as executor:
-                hashes = (
-                    value for typ, value in executor.map(worker, file_infos)
-                )
+                hashes = (hi.value for hi in executor.map(worker, file_infos))
                 return dict(zip(file_infos, hashes))
 
     def _collect_dir(self, path_info, **kwargs):
@@ -326,7 +310,7 @@ class BaseTree:
         new_hashes = self._calculate_hashes(not_in_state)
         hashes.update(new_hashes)
 
-        result = [
+        return [
             {
                 self.PARAM_CHECKSUM: hashes[fi],
                 # NOTE: this is lossy transformation:
@@ -342,37 +326,9 @@ class BaseTree:
             for fi in file_infos
         ]
 
-        # Sorting the list by path to ensure reproducibility
-        return sorted(result, key=itemgetter(self.PARAM_RELPATH))
-
-    def save_dir_info(self, dir_info):
-        typ, hash_, tmp_info = self._get_dir_info_hash(dir_info)
-        new_info = self.cache.tree.hash_to_path_info(hash_)
-        if self.cache.changed_cache_file(hash_):
-            self.cache.tree.makedirs(new_info.parent)
-            self.cache.tree.move(
-                tmp_info, new_info, mode=self.cache.CACHE_MODE
-            )
-
-        self.state.save(new_info, hash_)
-
-        return typ, hash_
-
-    def _get_dir_info_hash(self, dir_info):
-        # Sorting the list by path to ensure reproducibility
-        dir_info = sorted(dir_info, key=itemgetter(self.PARAM_RELPATH))
-
-        tmp = tempfile.NamedTemporaryFile(delete=False).name
-        with open(tmp, "w+") as fobj:
-            json.dump(dir_info, fobj, sort_keys=True)
-
-        tree = self.cache.tree
-        from_info = PathInfo(tmp)
-        to_info = tree.path_info / tmp_fname("")
-        tree.upload(from_info, to_info, no_progress_bar=True)
-
-        typ, hash_ = tree.get_file_hash(to_info)
-        return typ, hash_ + self.CHECKSUM_DIR_SUFFIX, to_info
+    def get_dir_hash(self, path_info, **kwargs):
+        dir_info = self._collect_dir(path_info, **kwargs)
+        return self.repo.cache.local.save_dir_info(dir_info)
 
     def upload(self, from_info, to_info, name=None, no_progress_bar=False):
         if not hasattr(self, "_upload"):
